# -*- coding: utf-8 -*-
#
#      Copyright (C) 2020 Axual B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import threading
from datetime import datetime
from time import sleep
from typing import List, Tuple, Callable

import confluent_kafka
from confluent_kafka import Consumer as KafkaConsumer
from confluent_kafka import TopicPartition, Message
from confluent_kafka.admin import ClusterMetadata

from axualclient.discovery import DiscoveryClient, DiscoveryClientRegistry, BOOTSTRAP_SERVERS_KEY, TIMESTAMP_KEY, \
    DISTRIBUTOR_TIMEOUT_KEY, DISTRIBUTOR_DISTANCE_KEY, TTL_KEY
from axualclient.patterns import resolve_group, resolve_topics
from axualclient.util import filter_axual_configuration, is_documented_by

logger = logging.getLogger(__name__)

DEFAULT_POLL_SPEED = 0.2


class Consumer(DiscoveryClient):
    """Simple balanced consumer class.
    Implements __iter__ to be able to create a for loop on the consumer
     to iterate through messages: for msg in Consumer. Set pause
     attribute to break from loop.
    Set poll_speed attribute to change the polling speed (default: 0.2 [secs])."""

    def __init__(self,
                 configuration: dict = None,
                 *args, **kwargs):
        """
        Instantiate a consumer for Axual. Derives from confluent_kafka
         Consumer class.
        Note that auto-commit is set to False, so received messages must
         be committed by your script's logic.

        Parameters
        ----------
        configuration: dict
            Configuration properties including Axual Configurations. All consumer Configurations can be found at:
             https://github.com/edenhill/librdkafka/blob/master/CONFIGURATION.md
        *args and **kwargs:
            Other parameters that can be passed to confluent_kafka Consumer.
        """
        self.unresolved_topics = []
        self.unresolved_group_id = configuration.get('application_id')
        self._consumer = None
        self.init_args = args
        self.init_kwargs = kwargs

        self.configuration = configuration
        # bootstrap servers & key/value serializers are not available at this point yet
        self.configuration['security.protocol'] = 'SSL'

        self.poll_speed = DEFAULT_POLL_SPEED
        self.initialized = False
        self.switch_lock = threading.Lock()
        self.init_lock = threading.Lock()

        self.discovery_result = {}
        self.discovery_fetcher = DiscoveryClientRegistry.register_client(
            self.configuration, self
        )

    def wait_for_initialization(self) -> None:
        if self.initialized:
            return
        with self.init_lock:
            self.discovery_fetcher.wait_for_discovery_result()

    def _do_with_switch_lock(self, func):
        self.wait_for_initialization()
        with self.switch_lock:
            return func()

    def _get_consumer(self):
        """ Convenience method added such that lambda expressions do not get evaluated on the old proxied object """
        return self._consumer

    def on_discovery_properties_changed(self, discovery_result: dict) -> None:
        """ A new discovery result has been received, need to switch """
        with self.switch_lock:
            self.discovery_result = discovery_result
            # plug in the new bootstrap servers
            self.configuration['bootstrap.servers'] = discovery_result[BOOTSTRAP_SERVERS_KEY]
            # plug in the resolved group.id
            self.configuration['group.id'] = resolve_group(discovery_result, self.unresolved_group_id)
            logger.debug(f'group.id: {self.configuration["group.id"]}')

            # Switch consumer
            if self.initialized:
                assignment = self._consumer.assignment()
                # sleep(self.poll_speed)  # Wait outstanding polls before closing consumer
                self._consumer.close()

                # Calculate switch time-out
                if len(assignment) > 0:
                    switch_timeout = self._calculate_switch_timeout(discovery_result)
                    sleep(switch_timeout / 1000)

            kafka_properties = filter_axual_configuration(self.configuration)
            self._consumer = KafkaConsumer(kafka_properties, *self.init_args, **self.init_kwargs)
            # subscribe to previously subscribed-to topics, on new cluster
            if self.unresolved_topics:
                resolved_topics = resolve_topics(self.discovery_result, self.unresolved_topics)
                self._consumer.subscribe(resolved_topics)
            self.initialized = True

    def _calculate_switch_timeout(self, discovery_result: dict):
        if self._is_at_least_once():
            return 0
        return max(int(discovery_result[DISTRIBUTOR_TIMEOUT_KEY]) *
                   int(discovery_result[DISTRIBUTOR_DISTANCE_KEY]) -
                   (datetime.utcnow() - discovery_result[TIMESTAMP_KEY]).total_seconds() * 1000,
                   int(discovery_result[TTL_KEY]))

    def _is_at_least_once(self) -> bool:
        return self.configuration.get('auto.offset.reset') in ['earliest', 'smallest', 'begin', 'start']

    def __iter__(self):
        """Continuously loop through messages until self.pause is set to True"""
        self.pause = False
        while not self.pause:
            msg = self.poll(self.poll_speed)
            yield msg

    # KafkaConsumer interface
    @is_documented_by(confluent_kafka.cimpl.Consumer.assign)
    def assign(self, partitions: List[TopicPartition], *args, **kwargs):
        return self._do_with_switch_lock(
            lambda: self._get_consumer().assign(partitions, *args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Consumer.assignment)
    def assignment(self, *args, **kwargs) -> List[TopicPartition]:
        return self._do_with_switch_lock(
            lambda: self._get_consumer().assignment(*args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Consumer.close)
    def close(self, *args, **kwargs) -> None:
        DiscoveryClientRegistry.deregister_client(self.unresolved_group_id, self)
        return self._do_with_switch_lock(
            lambda: self._get_consumer().close(*args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Consumer.commit)
    def commit(self,
               message: Message = None,
               offsets: List[TopicPartition] = None,
               asynchronous: bool = True,
               *args, **kwargs):
        if message is not None:
            kwargs['message'] = message
        if offsets is not None:
            kwargs['offsets'] = offsets
        kwargs['asynchronous'] = asynchronous
        return self._do_with_switch_lock(
            lambda: self._get_consumer().commit(*args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Consumer.committed)
    def committed(self,
                  partitions: List[TopicPartition],
                  timeout: float = None) -> List[TopicPartition]:
        if timeout is not None:
            return self._do_with_switch_lock(
                lambda: self._get_consumer().committed(partitions, timeout)
            )
        else:
            return self._do_with_switch_lock(
                lambda: self._get_consumer().committed(partitions)
            )

    @is_documented_by(confluent_kafka.cimpl.Consumer.consume)
    def consume(self,
                num_messages: int = 1,
                timeout: float = -1,
                *args, **kwargs) -> List[Message]:
        return self._do_with_switch_lock(
            lambda: self._get_consumer().consume(num_messages, timeout, *args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Consumer.consumer_group_metadata)
    def consumer_group_metadata(self):
        return self._do_with_switch_lock(
            lambda: self._get_consumer().consumer_group_metadata()
        )

    @is_documented_by(confluent_kafka.cimpl.Consumer.get_watermark_offsets)
    def get_watermark_offsets(self,
                              partition: TopicPartition,
                              timeout: float = None,
                              cached: bool = False,
                              *args, **kwargs) -> Tuple[int, int]:
        if timeout is not None:
            kwargs['timeout'] = timeout
        return self._do_with_switch_lock(
            lambda: self._get_consumer().get_watermark_offsets(partition=partition, cached=cached, *args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Consumer.list_topics)
    def list_topics(self,
                    topic: str = None,
                    timeout: float = -1,
                    *args, **kwargs) -> ClusterMetadata:
        if topic is not None:
            return self._do_with_switch_lock(
                lambda: self._get_consumer().list_topics(topic=topic, timeout=timeout, *args, **kwargs)
            )
        else:
            return self._do_with_switch_lock(
                lambda: self._get_consumer().list_topics(timeout=timeout, *args, **kwargs)
            )

    @is_documented_by(confluent_kafka.cimpl.Consumer.offsets_for_times)
    def offsets_for_times(self,
                          partitions: List[TopicPartition],
                          timeout: float = None) -> List[TopicPartition]:
        if timeout is not None:
            return self._do_with_switch_lock(
                lambda: self._get_consumer().offsets_for_times(partitions, timeout)
            )
        else:
            return self._do_with_switch_lock(
                lambda: self._get_consumer().offsets_for_times(partitions)
            )

    @is_documented_by(confluent_kafka.cimpl.Consumer.pause)
    def pause(self, partitions: List[TopicPartition]) -> None:
        return self._do_with_switch_lock(
            lambda: self._get_consumer().pause(partitions)
        )

    @is_documented_by(confluent_kafka.cimpl.Consumer.poll)
    def poll(self, timeout=None):
        if timeout is not None:
            return self._do_with_switch_lock(
                lambda: self._get_consumer().poll(timeout)
            )
        else:
            return self._do_with_switch_lock(
                lambda: self._get_consumer().poll()
            )

    @is_documented_by(confluent_kafka.cimpl.Consumer.position)
    def position(self, partitions: List[TopicPartition]) -> List[TopicPartition]:
        return self._do_with_switch_lock(
            lambda: self._get_consumer().position(partitions)
        )

    @is_documented_by(confluent_kafka.cimpl.Consumer.resume)
    def resume(self, partitions: List[TopicPartition]) -> None:
        return self._do_with_switch_lock(
            lambda: self._get_consumer().resume(partitions)
        )

    @is_documented_by(confluent_kafka.cimpl.Consumer.seek)
    def seek(self, partition: TopicPartition):
        return self._do_with_switch_lock(
            lambda: self._get_consumer().seek(partition)
        )

    @is_documented_by(confluent_kafka.cimpl.Consumer.store_offsets)
    def store_offsets(self,
                      message: confluent_kafka.Message = None,
                      offsets: List[TopicPartition] = None,
                      *args, **kwargs) -> None:
        if message is not None:
            kwargs['message'] = message
        if offsets is not None:
            kwargs['offsets'] = offsets
        return self._do_with_switch_lock(
            lambda: self._get_consumer().store_offsets(message, *args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Consumer.subscribe)
    def subscribe(self,
                  topics: List[str],
                  on_assign: Callable = None,
                  on_revoke: Callable = None,
                  on_lost: Callable = None,
                  *args, **kwargs):
        self.unresolved_topics = list(set(self.unresolved_topics + topics))
        if on_assign is not None:
            kwargs['on_assign'] = on_assign
        if on_revoke is not None:
            kwargs['on_revoke'] = on_revoke
        if on_lost is not None:
            kwargs['on_lost'] = on_lost
        return self._do_with_switch_lock(
            lambda: self._get_consumer().subscribe(
                resolve_topics(self.discovery_result, topics), *args, **kwargs
            )
        )

    @is_documented_by(confluent_kafka.cimpl.Consumer.unassign)
    def unassign(self, *args, **kwargs):
        return self._do_with_switch_lock(
            lambda: self._get_consumer().unassign(*args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Consumer.unsubscribe)
    def unsubscribe(self, *args, **kwargs):
        return self._do_with_switch_lock(
            lambda: self._get_consumer().unsubscribe(*args, **kwargs)
        )



# -*- coding: utf-8 -*-
#
#      Copyright (C) 2020 Axual B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import logging
from datetime import datetime

logger = logging.getLogger(__name__)


def dict_to_str(dictionary: dict) -> str:
    return json.dumps(dictionary, default=str, indent=2)


def filter_axual_configuration(configuration: dict) -> dict:
    """
    Returns a dict containing everything except for axual configuration from a dictionary
    """
    props = {}
    axual_configs = ['application_id', 'endpoint', 'environment', 'instance', 'tenant']
    for config in configuration:
        if config not in axual_configs:
            props[config] = configuration.get(config)
    return props


def first_of_string_list(comma_separated: str) -> str:
    return comma_separated.split(',')[0]


def calculate_producer_switch_timeout(is_keeping_order: bool,
                                      distributor_timeout_ms: int,
                                      distributor_distance: int,
                                      discovery_result_timestamp: datetime) -> float:
    """
    Returns the time needed for a producer to sleep prior to resuming produce operations
    @param is_keeping_order:           If not, no need to wait
    @param distributor_timeout_ms:     How long it takes for messages to distribute
    @param distributor_distance:       Defined in hops
    @param discovery_result_timestamp: When the discovery result was received
    @return: Time to sleep in seconds
    """
    if is_keeping_order:
        ms_since_discovery = max(
            (datetime.utcnow() - discovery_result_timestamp).total_seconds() * 1000,
            0
        )
        sleep_ms = distributor_timeout_ms * distributor_distance - ms_since_discovery
        return sleep_ms / 1000.
    return 0.


def is_documented_by(original: callable) -> callable:
    """
    Decorator for pointing to external documentation.
    @param original: component
    @return: callable
    """
    def wrapper(target):
        target.__doc__ = original.__doc__
        return target
    return wrapper



# -*- coding: utf-8 -*-
#
#      Copyright (C) 2021 Axual B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry.avro import AvroDeserializer as KafkaAvroDeserializer
from confluent_kafka.schema_registry.avro import AvroSerializer as KafkaAvroSerializer


class AvroSerializer(KafkaAvroSerializer):
    """ Configured on cluster detection/switch """

    def __init__(self,
                 schema_str: str,
                 to_dict: callable = None,
                 conf: dict = None
                 ):
        """
        @param schema_str (str): Avro `Schema Declaration. <https://avro.apache.org/docs/current/spec.html#schemas>`_
               If not provided, writer schema is used for deserialization.
        @param to_dict (callable, optional): Callable(object, SerializationContext) -> dict. Converts object to a dict.
        @param conf (dict): AvroSerializer configuration.
        """
        self._serializer = None
        self.schema_str = schema_str
        self.to_dict = to_dict
        self.conf = conf

        if self.conf is None:
            self.conf = {}
        self.conf['auto.register.schemas'] = False

    def __call__(self, value, ctx):
        if not self._serializer:
            raise RuntimeError('AvroSerializer is not configured')
        self._serializer.__call__(value, ctx)

    def configure(self, configuration: dict) -> None:
        """
        Create the AvroSerializer against the newly acquired SchemaRegistry url
        @param configuration: configuration to instantiate SchemaRegistryClient
        @return: None
        """
        schema_registry_client = SchemaRegistryClient(configuration)
        self._serializer = KafkaAvroSerializer(
            schema_registry_client=schema_registry_client,
            schema_str=self.schema_str,
            to_dict=self.to_dict,
            conf=self.conf
        )


class AvroDeserializer(KafkaAvroDeserializer):
    """ Configured on cluster detection/switch """

    def __init__(self,
                 schema_str: str = None,
                 from_dict: callable = None,
                 return_record_name: bool = False
                 ):
        """
        @param schema_str str (optional):  Avro reader schema declaration.
               If not provided, writer schema is used for deserialization.
        @param from_dict callable (optional):  Converts object literal(dict) to an instance of some object.
               must be callable with the signature: from_dict(SerializationContext, dict) -> object
        @param return_record_name (bool): If True, when reading a union of records, the result will
               be a tuple where the first value is the name of the record and the second value is
               the record itself.  Defaults to False.
        """
        self._deserializer = None
        self.schema_str = schema_str
        self.from_dict = from_dict
        self.return_record_name = return_record_name

    def __call__(self, value, ctx):
        if not self._deserializer:
            raise RuntimeError('AvroDeserializer is not configured')
        self._deserializer.__call__(value, ctx)

    def configure(self, configuration: dict) -> None:
        """
        Create the AvroDeserializer against the newly acquired SchemaRegistry url
        @param configuration: configuration to instantiate SchemaRegistryClient
        @return: None
        """
        schema_registry_client = SchemaRegistryClient(configuration)
        self._deserializer = KafkaAvroDeserializer(
            schema_registry_client=schema_registry_client,
            from_dict=self.from_dict,
            return_record_name=self.return_record_name
        )



# -*- coding: utf-8 -*-
#
#      Copyright (C) 2020 Axual B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import logging
from abc import ABC, abstractmethod
from datetime import datetime
from threading import Thread
from time import sleep

import requests
from requests import RequestException

from axualclient.exception import NoClusterAvailableException
from axualclient.util import dict_to_str
from axualclient.version import __version__

logger = logging.getLogger(__name__)


BACKOFF_PERIOD = 10  # seconds
V2_ENDPOINT_SUFFIX = 'v2'

# Discovery result keys
BOOTSTRAP_SERVERS_KEY = 'bootstrap.servers'
TIMESTAMP_KEY = 'timestamp'
TTL_KEY = 'ttl'
SCHEMA_REGISTRY_URL_KEY = 'schema.registry.url'
DISTRIBUTOR_TIMEOUT_KEY = 'distributor.timeout'
DISTRIBUTOR_DISTANCE_KEY = 'distributor.distance'


class DiscoveryClient(ABC):
    """ DiscoveryAPI observer interface, contains callback for change of properties. """

    @abstractmethod
    def on_discovery_properties_changed(self, discovery_result: dict):
        """ A new discovery result has been received. """
        pass


def _v2_suffix(endpoint: str) -> str:
    """ Appends `/v2` to the endpoint suffix. """
    version_two = V2_ENDPOINT_SUFFIX if endpoint.endswith('/') else '/' + V2_ENDPOINT_SUFFIX
    return endpoint + version_two


class DiscoveryFetcher:
    """ Makes calls against DiscoveryAPI to read the cluster assignment. """

    def __init__(self,
                 config: dict,
                 discovery_client: DiscoveryClient):
        self.endpoint = _v2_suffix(config['endpoint'])
        self.application_id = config['application_id']
        self.application_version = config.get('version', 'unknown')
        self.tenant = config['tenant']
        self.environment = config['environment']
        self.root_ca_location = config['ssl.ca.location']
        self.ssl_key_location = config['ssl.key.location']
        self.ssl_certificate_location = config['ssl.certificate.location']
        self.discovery_clients = [discovery_client]

        self.discovery_result = {}

        self.initialized = False
        self.stop_discovery = False
        # Serves as the termination condition if raised prior to a valid discovery response.
        self._initialization_exception = None
        self._discovery_thread = Thread(target=self._continuous_discovery, daemon=True)
        self._discovery_thread.name = 'discovery-' + self.application_id
        self._discovery_thread.setDaemon(True)
        self._discovery_thread.start()

    def close(self) -> None:
        logger.info('Stopping discovery thread.')
        self.stop_discovery = True

    def wait_for_discovery_result(self) -> None:
        while not self.discovery_result:
            if self._initialization_exception:
                raise self._initialization_exception
            logger.info('Waiting for Discovery Result.')
            sleep(BACKOFF_PERIOD)

    def register_discovery_client(self, client):
        self.discovery_clients.append(client)
        if self.discovery_result:
            client.on_discovery_properties_changed(self.discovery_result)

    def deregister_discovery_client(self, client):
        self.discovery_clients.remove(client)
        if len(self.discovery_clients) == 0:
            self.close()

    def _continuous_discovery(self) -> None:
        while not self.stop_discovery:
            self._execute_discovery_request()
            if not self.discovery_result or not self.discovery_result[BOOTSTRAP_SERVERS_KEY]:
                sleep(BACKOFF_PERIOD)
            else:
                sleep(int(self.discovery_result[TTL_KEY]) / 1000)

    def _create_headers(self) -> dict:
        return {
            'X-Application-Id': self.application_id,
            'X-Application-Version': self.application_version,
            'X-Client-Library': 'axual-python-client',
            'X-Client-Library-Version': __version__
        }

    def _create_params(self) -> dict:
        return {
            'applicationId': self.application_id,
            'tenant': self.tenant,
            'env': self.environment
        }

    def _execute_discovery_request(self) -> None:
        latest_properties = {}
        try:
            response = requests.get(url=self.endpoint,
                                    headers=self._create_headers(),
                                    params=self._create_params(),
                                    verify=self.root_ca_location,
                                    cert=(self.ssl_certificate_location, self.ssl_key_location),
                                    timeout=5)
            if response.ok:
                latest_properties = json.loads(response.content)
                latest_properties[TIMESTAMP_KEY] = datetime.utcnow()
                self._initialization_exception = None
            else:
                self._handle_non_200(response)
        except Exception as e:
            logger.error(f'Could not fetch Discovery API result: {e}')
            self._stop_if_not_initialized(e)
        self._process_discovery_result(latest_properties)

    def _handle_non_200(self, response) -> None:
        if response.status_code == 204:
            # Backoff should be triggered
            logger.warning('Empty response from Discovery API, no active clusters found')
            self._stop_if_not_initialized(NoClusterAvailableException)
        elif response.status_code == 403:
            logger.warning('Forbidden to query Discovery API')
            self._stop_if_not_initialized(NoClusterAvailableException)
        elif response.status_code == 503:
            logger.warning('Discovery API unavailable')
            self._stop_if_not_initialized(RequestException)
        else:
            msg = f'Unexpected response code from Discovery API: {response.status_code}'
            logger.warning(msg)
            self._stop_if_not_initialized(RequestException)

    def _stop_if_not_initialized(self, exception_type):
        if not self.initialized:
            self.stop_discovery = True
            self._initialization_exception = exception_type

    def _is_cluster_change_detected(self, discovery_result: dict) -> bool:
        new_bootstrap = discovery_result.get(BOOTSTRAP_SERVERS_KEY)
        return (not self.discovery_result and discovery_result) \
            or self.discovery_result.get(BOOTSTRAP_SERVERS_KEY) != new_bootstrap

    def _process_discovery_result(self, latest_discovery_result: dict) -> None:
        if latest_discovery_result and self._is_cluster_change_detected(latest_discovery_result):
            self.discovery_result = latest_discovery_result
            self.initialized = True
            logger.debug(f'Discovery result changed to: {dict_to_str(latest_discovery_result)}')
            for client in self.discovery_clients:
                client.on_discovery_properties_changed(latest_discovery_result)


class DiscoveryClientRegistry:
    discovery_clients = {}

    @classmethod
    def register_client(cls,
                        config: dict,
                        discovery_client: DiscoveryClient) -> DiscoveryFetcher:
        app_id = config['application_id']
        if app_id not in cls.discovery_clients:
            fetcher = DiscoveryFetcher(config, discovery_client)
            cls.discovery_clients[app_id] = fetcher
        else:
            cls.discovery_clients[app_id].register_discovery_client(discovery_client)
        return cls.discovery_clients[app_id]

    @classmethod
    def deregister_client(cls,
                          application_id: str,
                          discovery_client: DiscoveryClient) -> None:
        if application_id in cls.discovery_clients:
            cls.discovery_clients[application_id]\
                .deregister_discovery_client(discovery_client)



# -*- coding: utf-8 -*-
#
#      Copyright (C) 2021 Axual B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional, List

GROUP_PLACEHOLDER = '{group}'
ENVIRONMENT_PLACEHOLDER = '{environment}'
INSTANCE_PLACEHOLDER = '{instance}'
TENANT_PLACEHOLDER = '{tenant}'
TOPIC_PLACEHOLDER = '{topic}'

# Move those on another class maybe
GROUP_KEY = 'group'
ENVIRONMENT_KEY = 'environment'
INSTANCE_KEY = 'instance'
TENANT_KEY = 'tenant'
TOPIC_KEY = 'topic'
TOPIC_PATTERN_KEY = 'topic.pattern'
GROUP_PATTERN_KEY = 'group.id.pattern'


def resolve_topic(discovery_result: dict, topic: Optional[str]) -> Optional[str]:
    if topic is None:
        return
    return discovery_result[TOPIC_PATTERN_KEY] \
        .replace(TENANT_PLACEHOLDER, discovery_result[TENANT_KEY]) \
        .replace(INSTANCE_PLACEHOLDER, discovery_result[INSTANCE_KEY]) \
        .replace(ENVIRONMENT_PLACEHOLDER, discovery_result[ENVIRONMENT_KEY]) \
        .replace(TOPIC_PLACEHOLDER, topic)


def resolve_topics(discovery_result: dict, topics: Optional[list]) -> Optional[List[str]]:
    if not topics:
        return None
    return [resolve_topic(discovery_result, t) for t in topics]


def resolve_group(discovery_result: dict, group: Optional[str]) -> Optional[str]:
    if group is None:
        return
    return discovery_result[GROUP_PATTERN_KEY] \
        .replace(TENANT_PLACEHOLDER, discovery_result[TENANT_KEY]) \
        .replace(INSTANCE_PLACEHOLDER, discovery_result[INSTANCE_KEY]) \
        .replace(ENVIRONMENT_PLACEHOLDER, discovery_result[ENVIRONMENT_KEY]) \
        .replace(GROUP_PLACEHOLDER, group)



# -*- coding: utf-8 -*-
#
#      Copyright (C) 2020 Axual B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import threading
from time import sleep
from typing import Callable
from typing import List

import confluent_kafka
from confluent_kafka import Producer as KafkaProducer, TopicPartition
from confluent_kafka.admin import ClusterMetadata

from axualclient.discovery import DiscoveryClientRegistry, DiscoveryClient, BOOTSTRAP_SERVERS_KEY, TIMESTAMP_KEY, \
    DISTRIBUTOR_TIMEOUT_KEY, DISTRIBUTOR_DISTANCE_KEY
from axualclient.patterns import resolve_topic
from axualclient.util import filter_axual_configuration, calculate_producer_switch_timeout, is_documented_by

logger = logging.getLogger(__name__)


class Producer(DiscoveryClient):

    def __init__(self,
                 configuration: dict,
                 *args, **kwargs):
        """
        Instantiate a producer for Axual. The _producer attribute is the confluent_kafka Producer class.

        Parameters
        ----------
        configuration : dict
            Configuration properties including Axual Configuration. All producer Configuration can be found at:
             https://github.com/edenhill/librdkafka/blob/master/CONFIGURATION.md
        *args and **kwargs :
            Other parameters that can be passed to confluent_kafka Producer.
        """
        self._producer = None  # no discovery result yet

        self.init_args = args
        self.init_kwargs = kwargs

        self.configuration = configuration
        # bootstrap servers & key/value serializers are not available at this point yet
        self.configuration['security.protocol'] = 'SSL'
        if 'partitioner' not in self.configuration.keys():
            self.configuration['partitioner'] = 'murmur2_random'

        self.switch_lock = threading.Lock()
        self.init_lock = threading.Lock()

        self.discovery_result = {}
        self.discovery_fetcher = DiscoveryClientRegistry.register_client(
            self.configuration, self
        )

    def wait_for_initialization(self) -> None:
        if self._producer is not None:
            return
        with self.init_lock:
            self.discovery_fetcher.wait_for_discovery_result()

    def _do_with_switch_lock(self, func):
        self.wait_for_initialization()
        with self.switch_lock:
            return func()

    def _get_producer(self):
        """ Convenience method added such that lambda expressions do not get evaluated on the old proxied object """
        return self._producer

    def on_discovery_properties_changed(self, discovery_result: dict) -> None:
        """ A new discovery result has been received, need to switch """
        with self.switch_lock:
            self.discovery_result = discovery_result
            # plug in the new bootstrap servers
            self.configuration['bootstrap.servers'] = discovery_result[BOOTSTRAP_SERVERS_KEY]

            # Switch producer
            if self._producer is not None:
                self._producer.flush()
                switch_timeout = calculate_producer_switch_timeout(
                    self._is_keeping_order(),
                    int(discovery_result[DISTRIBUTOR_TIMEOUT_KEY]),
                    int(discovery_result[DISTRIBUTOR_DISTANCE_KEY]),
                    discovery_result[TIMESTAMP_KEY]
                )
                sleep(switch_timeout)

            kafka_properties = filter_axual_configuration(self.configuration)
            self._producer = KafkaProducer(kafka_properties, *self.init_args, **self.init_kwargs)

    def _is_keeping_order(self):
        return self.configuration.get('max.in.flight.requests.per.connection') in [1, '1']

    @is_documented_by(confluent_kafka.cimpl.Producer.produce)
    def produce(self,
                topic: str,
                value=None,
                key=None,
                partition: int = None,
                on_delivery: Callable = None,
                timestamp: int = None,
                headers: dict = None,
                *args, **kwargs) -> None:
        if value is not None:
            kwargs['value'] = value
        if key is not None:
            kwargs['key'] = key
        if partition is not None:
            kwargs['partition'] = partition
        if on_delivery is not None:
            kwargs['on_delivery'] = on_delivery
        if timestamp is not None:
            kwargs['timestamp'] = timestamp
        if headers is not None:
            kwargs['headers'] = headers
        self._do_with_switch_lock(
            lambda:
            self._get_producer().produce(resolve_topic(self.discovery_result, topic), *args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Producer.abort_transaction)
    def abort_transaction(self, timeout: float = None, *args, **kwargs):
        if timeout is not None:
            self._do_with_switch_lock(
                lambda: self._get_producer().abort_transaction(timeout, *args, **kwargs)
            )
        else :
            self._do_with_switch_lock(
                lambda: self._get_producer().abort_transaction(*args, **kwargs)
            )

    @is_documented_by(confluent_kafka.cimpl.Producer.begin_transaction)
    def begin_transaction(self, *args, **kwargs):
        self._do_with_switch_lock(
            lambda: self._get_producer().begin_transaction(*args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Producer.commit_transaction)
    def commit_transaction(self, timeout: float = None, *args, **kwargs):
        if timeout is not None:
            self._do_with_switch_lock(
                lambda: self._get_producer().commit_transaction(timeout, *args, **kwargs)
            )
        else:
            self._do_with_switch_lock(
                lambda: self._get_producer().commit_transaction(*args, **kwargs)
            )

    @is_documented_by(confluent_kafka.cimpl.Producer.flush)
    def flush(self, timeout: float = None, *args, **kwargs) -> int:
        if timeout is None:
            return self._do_with_switch_lock(
                lambda: self._get_producer().flush(*args, **kwargs)
            )
        return self._do_with_switch_lock(
            lambda: self._get_producer().flush(timeout, *args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Producer.init_transactions)
    def init_transactions(self, timeout: float, *args, **kwargs) -> None:
        self._do_with_switch_lock(
            lambda: self._get_producer().init_transactions(timeout, *args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Producer.list_topics)
    def list_topics(self,
                    topic: str = None,
                    timeout: float = -1,
                    *args, **kwargs) -> ClusterMetadata:
        if topic is not None:
            return self._do_with_switch_lock(
                lambda: self._get_producer().list_topics(topic, timeout, *args, **kwargs)
            )
        else:
            return self._do_with_switch_lock(
                lambda: self._get_producer().list_topics(timeout, *args, **kwargs)
            )

    @is_documented_by(confluent_kafka.cimpl.Producer.poll)
    def poll(self, timeout=None, *args, **kwargs):
        if timeout is not None:
            kwargs['timeout'] = timeout
        return self._do_with_switch_lock(
            lambda: self._get_producer().poll(*args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Producer.purge)
    def purge(self, in_queue: bool = True, in_flight: bool = True, blocking: bool = True, *args, **kwargs):
        return self._do_with_switch_lock(
            lambda: self._get_producer().purge(in_queue, in_flight, blocking, *args, **kwargs)
        )

    @is_documented_by(confluent_kafka.cimpl.Producer.send_offsets_to_transaction)
    def send_offsets_to_transaction(self,
                                    offsets: List[TopicPartition],
                                    group_metadata: object,
                                    timeout=None,
                                    *args, **kwargs):
        if timeout is not None:
            kwargs['timeout'] = timeout
        return self._do_with_switch_lock(
            lambda: self._get_producer().send_offsets_to_transaction(offsets, group_metadata, *args, **kwargs)
        )


    def __class__(self, *args, **kwargs):
        return self._producer.__class__(*args, **kwargs)

    def __dir__(self, *args, **kwargs):
        return self._producer.__dir__(*args, **kwargs)

    def __doc__(self, *args, **kwargs):
        return self._producer.__doc__(*args, **kwargs)

    def __eq__(self, *args, **kwargs):
        return self._producer.__eq__(*args, **kwargs)

    def __format__(self, *args, **kwargs):
        return self._producer.__format__(*args, **kwargs)

    def __ge__(self, *args, **kwargs):
        return self._producer.__ge__(*args, **kwargs)

    def __gt__(self, *args, **kwargs):
        return self._producer.__gt__(*args, **kwargs)

    def __hash__(self, *args, **kwargs):
        return self._producer.__hash__(*args, **kwargs)

    def __le__(self, *args, **kwargs):
        return self._producer.__le__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        return self._producer.__len__(*args, **kwargs)

    def __lt__(self, *args, **kwargs):
        return self._producer.__lt__(*args, **kwargs)

    def __ne__(self, *args, **kwargs):
        return self._producer.__ne__(*args, **kwargs)

    def __reduce__(self, *args, **kwargs):
        return self._producer.__reduce__(*args, **kwargs)

    def __reduce_ex__(self, *args, **kwargs):
        return self._producer.__reduce_ex__(*args, **kwargs)

    def __repr__(self, *args, **kwargs):
        return self._producer.__repr__(*args, **kwargs)

    def __sizeof__(self, *args, **kwargs):
        return self._producer.__sizeof__(*args, **kwargs)

    def __str__(self, *args, **kwargs):
        return self._producer.__str__(*args, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['axualclient']

package_data = \
{'': ['*']}

install_requires = \
['confluent-kafka[avro]>=1.7.0,<2.0.0']

setup_kwargs = {
    'name': 'axual-client-python',
    'version': '1.1.0',
    'description': 'Axual client for Python',
    'long_description': "Axual Python Client\n-------------------\n\n[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://opensource.org/licenses/Apache-2.0)\n[![Pipeline Status](https://gitlab.com/axual-public/axual-client-python/badges/master/pipeline.svg)](https://gitlab.com/axual-public/axual-client-python/commits/master) \n[![Quality Gate Status](https://sonarcloud.io/api/project_badges/measure?project=axual-public-axual-client-python&metric=alert_status&token=803308122363db3988c6a86bfedcb42d5b4a6791)](https://sonarcloud.io/dashboard?id=axual-public-axual-client-python)\n[![Maintainability Rating](https://sonarcloud.io/api/project_badges/measure?project=axual-public-axual-client-python&metric=sqale_rating&token=803308122363db3988c6a86bfedcb42d5b4a6791)](https://sonarcloud.io/dashboard?id=axual-public-axual-client-python)\n\nPython client library that enables interaction with the Axual platform.\n\n![Python Client Overview](http://www.plantuml.com/plantuml/png/VP0zQyCm48Rdw5TSJI6J6-hP14AxT0jqB1s4T8v5zA7ALca8_Uzrpht4zf39kkTz_HpSW_7APerGAnkoDhupXxRV76Lpb5iXrc8DAaI36eotnYEqc12Q51pqBKAqqlVPsGqqmMfCJyCVeadI8HJx57HMc436M83i838uYvKABWQFleTYzvEb1MNDC5rxuoX-MVOIV9VTHAO8t9TnxofZ6xKebjG_YvdarNGgV6CwmMx_HZLz8RFEkshHIKXuC5sVJjJYGQo-CcQ4eduS3qafFP_lP9KGAre4vKThj4R_MMudKdxDOhQENoWPLV-eWlBMkfpWTbseXuGMBY5lMd3MS4niSc1oBI4z5uhg3m00)\n\n## Prerequisites\n\nPython 3 is required, Python 3.6 or greater is recommended.\n\n## Installation\n\n```bash\npip install axual-client-python\n```\n\n## Testing\nTests are located in the `tests/` directory.\nTo run all of them:\n```bash\npython -m unittest discover tests\n```\n\n## Usage\n### Producer\n\n```python\nimport json\nfrom axualclient.producer import Producer\n\nconf = {\n    # Axual configuration\n    'application_id': application_id,\n    'endpoint': endpoint,\n    'tenant': tenant,\n    'environment': environment,\n    # SSL configuration\n    'ssl.certificate.location': producer_app_cert,\n    'ssl.key.location': producer_app_key,\n    'ssl.ca.location': root_cert,\n}\nproducer = Producer(conf)\nproducer.produce(value=json.dumps(dict(a=1, b='banana')))\nproducer.flush()  # Flushes producer before ending (triggering callbacks for delivery reports)\n```\nSee [SSL Configuration](#ssl-configuration) for more info on the certificates required.\n\n### Consumer\n\n```python\nfrom axualclient.consumer import Consumer\n\nconf = {\n    # Axual configuration\n    'application_id': application_id,\n    'endpoint': endpoint,\n    'tenant': tenant,\n    'environment': environment,\n    # SSL configuration\n    'ssl.certificate.location': producer_app_cert,\n    'ssl.key.location': producer_app_key,\n    'ssl.ca.location': root_cert,\n    # Consumer configuration\n    'auto.offset.reset': 'earliest',\n    'on_commit': on_commit_callback,\n}\nconsumer = Consumer(conf)\n\ntry:\n    # Subscribe to topic(s) as they appear on Self Service\n    consumer.subscribe(['your-topic'])\n    while True:\n        msg = consumer.poll(1.0)\n\n        # msg is None if no new message\n        if msg is None:\n            continue\n\n        if msg.error():\n            # Error handling\n            raise KafkaException(msg.error())\n        else:\n            # Do some processing on the message\n            print(\n                f'Received message on topic {msg.topic()} partition {msg.partition()} '\n                f'offset {msg.offset()} key {msg.key()} value {msg.value()}'\n            )\n            # Commit message offset to the topic\n            consumer.commit()\nfinally:\n    if consumer is not None:\n        # Cleanly unregister from cluster by closing consumer\n        consumer.commit()\n        consumer.close()\n```\n\n### Producer (AVRO)\nProducing AVRO messages works if you have the AVRO schema available that was uploaded to self-service. \nThose schemas need to be provided to the serializer via the `schema_str` as Avro [Schema Declaration](https://avro.apache.org/docs/current/spec.html#schemas>).\n```python\nfrom axualclient.avro import AvroSerializer\nfrom axualclient.serializing_producer import SerializingProducer\n\nkey_serializer = AvroSerializer(schema_str=Application.SCHEMA, to_dict=application_to_dict)\nvalue_serializer = AvroSerializer(schema_str=ApplicationLogEvent.SCHEMA, to_dict=application_log_event_to_dict)\n\nconf = {\n    # Axual configuration\n    'application_id': application_id,\n    'endpoint': endpoint,\n    'tenant': tenant,\n    'environment': environment,\n    # SSL configuration\n    'ssl.certificate.location': producer_app_cert,\n    'ssl.key.location': producer_app_key,\n    'ssl.ca.location': root_cert,\n    # Producer configuration\n    'key.serializer': key_serializer,\n    'value.serializer': value_serializer,\n    'acks': 'all',\n}\nproducer = SerializingProducer(conf)\n\nproducer.produce(topic=topic, value=value, key=key)\nproducer.flush() # Flushes producer before ending (triggering callbacks for delivery reports)\n```\n\n### Consumer (AVRO)\nConsumer `AvroDeserializers` need to be instantiated and passed as configuration to the Axual `DeserializingConsumer` like so:\n```python\nfrom axualclient.avro import AvroDeserializer\nfrom axualclient.deserializing_consumer import DeserializingConsumer\n\n# Initialize Deserializers:\nkey_deserializer = AvroDeserializer(\n    # Optional parameters\n    schema_str=Application.SCHEMA,\n    from_dict=dict_to_application\n)\nvalue_deserializer = AvroDeserializer(\n    # Optional parameters\n    schema_str=ApplicationLogEvent.SCHEMA,\n    from_dict=dict_to_application_log_event,\n)\n\nconfiguration = {\n    # Axual configuration\n    'application_id': application_id,\n    'endpoint': endpoint,\n    'tenant': tenant,\n    'environment': environment,\n    # SSL configuration\n    'ssl.certificate.location': producer_app_cert,\n    'ssl.key.location': producer_app_key,\n    'ssl.ca.location': root_cert,\n    # Consumer configuration\n    'key.deserializer': key_deserializer,\n    'value.deserializer': value_deserializer,\n    'auto.offset.reset': 'earliest',\n    'on_commit': on_commit_callback,\n    'error_cb': on_error_callback,\n    'logger': logger\n}\n\nconsumer = DeserializingConsumer(configuration)\n\ntry:\n    consumer.subscribe([topic])\n    while True:\n        msg = consumer.poll()\n\n        if msg is None:\n            continue\n\n        if msg.error():\n            print(f'Error returned by poll: {msg.error()}')\n        else:\n            print(\n                f'Received message on topic {msg.topic()} partition {msg.partition()} '\n                f'offset {msg.offset()} key: {str(msg.key())} value: {str(msg.value())}'\n            )\n            consumer.commit()\nexcept KeyboardInterrupt:\n    print('Caught KeyboardInterrupt, stopping.')\nfinally:\n    if consumer is not None:\n        print('Committing final offsets and leaving group.')\n        consumer.commit()\n        consumer.close()\n```\n\n\n## SSL Configuration\nThe client configuration requires a correct SSL configuration in order to communicate securely with brokers, the discovery API, and the schema registry.  \nEach application (as defined in self-service) requires an application certificate (`ssl.certificate.location`) and corresponding private key (`ssl.key.location`). \nThe application certificate must match the one uploaded in self-service for that application.  \nThe file with root certificates needs to be created properly: the brokers might be using a root certificate authority different \nfrom the authority that signed the certificates for the discovery API and schema registry.\nThe base64-encoded unencrypted versions of these certificates can be pasted into one file (`ssl.ca.location`).\nThis file would then look like the following example:\n```\n-----BEGIN CERTIFICATE-----\nMIIQdaGDAksKadksSDKNsdka5sjy8elAMsm3d .....\n ...  more base64-encoded content here ...\n..... LKlmsf02mz2EWYnds=\n-----END CERTIFICATE-----\n\n-----BEGIN CERTIFICATE-----\nMIIPmsd92nNWlasHWdwMOe92nwoa2QNinnNaZ .....\n ...  more base64-encoded content here ...\n..... ldFAf02SArubBW7wVFW2i1=\n-----END CERTIFICATE-----\n```\n\n## Examples\n\nSimple use cases using the client code can be found in the \n[Axual Python Client Examples](https://gitlab.com/axual-public/axual-client-python-examples).\n\n\n## Known Limitations\n\n- We have encountered issues when using `root_ca`s consisting of more than 1 intermediate certificates.\nThe issue originates with the underlying SSL C library implementation and results in the following exception \nwhen authenticating:\n\n```\n[SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed: path length constraint exceeded (_ssl.c:1123)\n```\n- Transaction support is not there. An [issue](https://gitlab.com/axual-public/axual-client-python/-/issues/17) is already created and transactions will be supported in future releases.\n- PEM as string format is not supported. An [issue](https://gitlab.com/axual-public/axual-client-python/-/issues/19) is already created and will be added in future releases.\n\n## Contributing\n\nAxual is interested in building the community; we would welcome any thoughts or \n[patches](https://gitlab.com/axual-public/axual-client-python/-/issues).\nYou can reach us [here](https://axual.com/contact/).\n\nSee [contributing](https://gitlab.com/axual-public/axual-client-python/blob/master/CONTRIBUTING.md).\n\n### Building locally\nThis project uses [poetry](https://python-poetry.org/docs/) for dependency management and building.\nInstall the tool as per the instructions on the linked page, and build the library using:\n\n    poetry build\n\n#### For maintainers: building a release\nThe version of the library being built will be the version specified in `pyproject.toml` whenever you \npush to a branch. \n<br>\nWhen tagging and building a release however, please be aware that the CI pipeline will ignore the version in `pyproject.toml` and \nbuild a release based on what is specified in the tag; \nfor example tagging `1.0.0-alpha4` will produce `axual_client_python-1.0.0a4-py3-none-any.whl`.\n<p>\nThis has two consequences:\n\n- You have to follow the normal [semver](https://semver.org/) rules when choosing a tag\n- After releasing, it falls on the developer to manually update the version in `pyproject.toml` in preparation for the next version.\n\n\n## License\n\nAxual Python Client is licensed under the [Apache License, Version 2.0](https://www.apache.org/licenses/LICENSE-2.0.txt).",
    'author': 'Axual',
    'author_email': 'opensource@axual.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
